float4x4 tW : WORLD;
float4x4 tWV: WORLDVIEW ;
float4x4 tWVP : WORLDVIEWPROJECTION ;
float4x4 tWIT: WORLDINVERSETRANSPOSE ;

float4x4 tV : VIEW ;
float4x4 tVI : VIEWINVERSE ;
float4x4 tP : PROJECTION ;
float4x4 tVP : VIEWPROJECTION ;
float3 cam_Pos : CAMERAPOSITION;

bool FrustumCulling;
float VolumeTextureAmount = 1;
float VolTexPhase = 0;
float VolTexPower = 1;
float VolTexUVScale = 1;

float4 cAmb <bool color=true;String uiname="Color";> = { 1.0f,1.0f,1.0f,1.0f };
float4x4 tProjectionDepth;

StructuredBuffer<float4> PosScaleBuffer;
StructuredBuffer<float3> Random3dBuffer;

float2 sizeCloseupFadeOut_minMaxDist = 1;
float closeFade;
float tau = 1;

float gVelocityGain : GLOBAL_MOTIONBLUR_MULT;

float2 Rxy : TARGETSIZE;


Texture2D LinearDepthMap;
Texture3D NoiseVolume;

SamplerState pointSampler : IMMUTABLE
{
    Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
};
SamplerState linearSampler : IMMUTABLE
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
	AddressW = Wrap;
};

//==============================================================================
// FUNCTIONS ===================================================================
//==============================================================================

float Culling(float3 pos)
{
	float4 projected = mul(float4(pos, 1), tVP);
	projected.xyz /= projected.w;
	float result = 0;
	
	if(	all(projected.xy > -1.05) && 
		all(projected.xy < 1.05) && 
		projected.z > 0 && 
		projected.z < 1)
	{	result = 1;	}
	return result;
}

//==============================================================================
// VERTEX SHADER ===============================================================
//==============================================================================

struct VS_IN
{
	uint iv : SV_VertexID;
};

struct VS_OUT
{
    float3 posW: POSITION ;	
	float4 col : COLOR ;
	float3 info : TEXCOORD1 ;
	float size : TEXCOORD2 ;
	float3 vel : TEXCOORD3 ;
	uint iv : TEXCOORD4;
};

VS_OUT VS(VS_IN In)
{
    VS_OUT Out = (VS_OUT)0;
	
	uint id = In.iv;
	Out.iv = id;
	Out.col = cAmb;
	
    Out.posW = mul(float4(PosScaleBuffer[id].xyz,1), tW).xyz;
	
	Out.size = PosScaleBuffer[id].w;
	
	
	
	
    return Out;
}

//==============================================================================



//==============================================================================
// GEOMETRY SHADER =============================================================
//==============================================================================

float3x3 lookat(float3 dir,float3 up=float3(0,1,0)){float3 z=normalize(dir);float3 x=normalize(cross(up,z));float3 y=normalize(cross(z,x));return float3x3(x,y,z);} 

float3 g_positions[4]:IMMUTABLE =
{
    float3( -1, 1, 0 ),
    float3( 1, 1, 0 ),
    float3( -1, -1, 0 ),
    float3( 1, -1, 0 ),
};
float2 g_texcoords[4]:IMMUTABLE = 
{ 
    float2(0,1), 
    float2(1,1),
    float2(0,0),
    float2(1,0),
};

struct GS_OUT
{
    float4 posWVP: SV_POSITION ;
	float3 posW: POSITION1 ;
	float3 posWV: POSITION3 ;
	float4 col : COLOR ;
	float3 texCd : TEXCOORD1 ;
	float size : TEXCOORD2 ;
	float3x3 rot : TEXCOORD3 ;
	float3 PPosV : POSITION4;
};

[maxvertexcount(4)]
void GS(point VS_OUT In[1], inout TriangleStream<GS_OUT> SpriteStream)
{
    GS_OUT Out;
    
	//bool enable = Culling(In[0].posW);
	if(Culling(In[0].posW) || !FrustumCulling)
	{		
	    Out.col = In[0].col;
		Out.size = In[0].size;
		
		Out.posW = In[0].posW.xyz;
		
		Out.rot = lookat(mul(float4(In[0].posW.xyz-tVI[3].xyz, 0), tV));
		
		Out.PPosV = mul(float4(In[0].posW,1.0), tV).xyz;
		
		
		uint id = In[0].iv;
		
		
		

		for(int i=0; i<4; i++)
		{
			float3 position = g_positions[i] * In[0].size;
		    position = mul(position, (float3x3)tVI ) + In[0].posW.xyz;
			Out.posWV = mul(float4(position,1.0), tV).xyz;
		    Out.posWVP = mul(float4(position,1.0), tVP);
			
		    Out.texCd = float3(g_texcoords[i],0);	
			Out.texCd.z = VolTexPhase;
			Out.texCd *= VolTexUVScale;
			Out.texCd += Random3dBuffer[id];
		    SpriteStream.Append(Out);
		}
		
		
		
		SpriteStream.RestartStrip();
		
	}
}

//==============================================================================




//==============================================================================
// PIXEL SHADER ================================================================
//==============================================================================


float4 PS_constant(GS_OUT In) : SV_Target
{
	
	//clip(1-length(In.texCd-.5)-0.5f);
	
	float4 col = In.col;
	
	
    return col;
	
}




float SoftOpacity(float3 P, float3 Q, float r, float depthVal, float f, float tau) {

	float alpha = 0;
	float d = length(P.xy - Q.xy);
		if(d < r) {
		float w = sqrt(r*r - d*d);
		float F = P.z - w;
		float B = P.z + w;
		float ds = min(depthVal, B) - max(f, F);
		alpha = 1 - exp(-tau * (1-d/r) * ds);
	}
	return alpha;
}

float linearDepth(float d)
{
	return tProjectionDepth._43 / (d - tProjectionDepth._33);
}

float4 PS_soft(GS_OUT In) : SV_Target
{
	
	float texVal = NoiseVolume.Sample(linearSampler,In.texCd).r;
	
	//clip(1-length(In.texCd-.5)-0.5f);
	
	float2 screenUV = In.posWVP.xy / Rxy;

	float depth = LinearDepthMap.Sample(pointSampler,screenUV).r;

	
	float4 col = In.col;
	float mask = SoftOpacity(In.posWV, In.PPosV, In.size, depth, closeFade, tau);
	mask = saturate(mask);
	col.a *= mask;
	
    return col * lerp(1,pow(texVal,VolTexPower),VolumeTextureAmount);
	
}



//==============================================================================
// TECHIQUES ===================================================================
//==============================================================================

technique11 Constant
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_5_0, VS() ) );
		SetGeometryShader( CompileShader( gs_5_0, GS() ) );
		SetPixelShader( CompileShader( ps_5_0, PS_constant() ) );
	}
}

technique11 Soft
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_5_0, VS() ) );
		SetGeometryShader( CompileShader( gs_5_0, GS() ) );
		SetPixelShader( CompileShader( ps_5_0, PS_soft() ) );
	}
}